@@ -1,33 +1,27 @@
 import string
 from types import MappingProxyType
 from typing import (
-    TYPE_CHECKING,
+    IO,
     Any,
     Callable,
     Dict,
     FrozenSet,
     Iterable,
+    NamedTuple,
     Optional,
-    TextIO,
     Tuple,
 )
+import warnings
 
 from tomli._re import (
-    RE_BIN,
     RE_DATETIME,
-    RE_HEX,
     RE_LOCALTIME,
     RE_NUMBER,
-    RE_OCT,
     match_to_datetime,
     match_to_localtime,
     match_to_number,
 )
 
-if TYPE_CHECKING:
-    from re import Pattern
-
-
 ASCII_CTRL = frozenset(chr(i) for i in range(32)) | frozenset(chr(127))
 
 # Neither of these sets include quotation mark or backslash. They are
@@ -44,6 +38,7 @@ TOML_WS = frozenset(" \t")
 TOML_WS_AND_NEWLINE = TOML_WS | frozenset("\n")
 BARE_KEY_CHARS = frozenset(string.ascii_letters + string.digits + "-_")
 KEY_INITIAL_CHARS = BARE_KEY_CHARS | frozenset("\"'")
+HEXDIGIT_CHARS = frozenset(string.hexdigits)
 
 BASIC_STR_ESCAPE_REPLACEMENTS = MappingProxyType(
     {
@@ -67,9 +62,17 @@ class TOMLDecodeError(ValueError):
     """An error raised if a document is not valid TOML."""
 
 
-def load(fp: TextIO, *, parse_float: ParseFloat = float) -> Dict[str, Any]:
+def load(fp: IO, *, parse_float: ParseFloat = float) -> Dict[str, Any]:
     """Parse TOML from a file object."""
     s = fp.read()
+    if isinstance(s, bytes):
+        s = s.decode()
+    else:
+        warnings.warn(
+            "Text file object support is deprecated in favor of binary file objects."
+            ' Use `open("foo.toml", "rb")` to open the file in binary mode.',
+            DeprecationWarning,
+        )
     return loads(s, parse_float=parse_float)
 
 
@@ -80,7 +83,8 @@ def loads(s: str, *, parse_float: ParseFloat = float) -> Dict[str, Any]:  # noqa
     # literals. Let's do so to simplify parsing.
     src = s.replace("\r\n", "\n")
     pos = 0
-    state = State()
+    out = Output(NestedDict(), Flags())
+    header: Key = ()
 
     # Parse one statement at a time
     # (typically means one line in TOML source)
@@ -104,7 +108,7 @@ def loads(s: str, *, parse_float: ParseFloat = float) -> Dict[str, Any]:  # noqa
             pos += 1
             continue
         if char in KEY_INITIAL_CHARS:
-            pos = key_value_rule(src, pos, state, parse_float)
+            pos = key_value_rule(src, pos, out, header, parse_float)
             pos = skip_chars(src, pos, TOML_WS)
         elif char == "[":
             try:
@@ -112,9 +116,9 @@ def loads(s: str, *, parse_float: ParseFloat = float) -> Dict[str, Any]:  # noqa
             except IndexError:
                 second_char = None
             if second_char == "[":
-                pos = create_list_rule(src, pos, state)
+                pos, header = create_list_rule(src, pos, out)
             else:
-                pos = create_dict_rule(src, pos, state)
+                pos, header = create_dict_rule(src, pos, out)
             pos = skip_chars(src, pos, TOML_WS)
         elif char != "#":
             raise suffixed_err(src, pos, "Invalid statement")
@@ -133,17 +137,7 @@ def loads(s: str, *, parse_float: ParseFloat = float) -> Dict[str, Any]:  # noqa
             )
         pos += 1
 
-    return state.out.dict
-
-
-class State:
-    def __init__(self) -> None:
-        # Mutable, read-only
-        self.out = NestedDict()
-        self.flags = Flags()
-
-        # Immutable, read and write
-        self.header_namespace: Key = ()
+    return out.data.dict
 
 
 class Flags:
@@ -242,6 +236,11 @@ class NestedDict:
             cont[last_key] = [{}]
 
 
+class Output(NamedTuple):
+    data: NestedDict
+    flags: Flags
+
+
 def skip_chars(src: str, pos: Pos, chars: Iterable[str]) -> Pos:
     try:
         while src[pos] in chars:
@@ -266,11 +265,10 @@ def skip_until(
         if error_on_eof:
             raise suffixed_err(src, new_pos, f'Expected "{expect!r}"')
 
-    bad_chars = error_on.intersection(src[pos:new_pos])
-    if bad_chars:
-        bad_char = next(iter(bad_chars))
-        bad_pos = src.index(bad_char, pos)
-        raise suffixed_err(src, bad_pos, f'Found invalid character "{bad_char!r}"')
+    if not error_on.isdisjoint(src[pos:new_pos]):
+        while src[pos] not in error_on:
+            pos += 1
+        raise suffixed_err(src, pos, f'Found invalid character "{src[pos]!r}"')
     return new_pos
 
 
@@ -295,74 +293,67 @@ def skip_comments_and_array_ws(src: str, pos: Pos) -> Pos:
             return pos
 
 
-def create_dict_rule(src: str, pos: Pos, state: State) -> Pos:
+def create_dict_rule(src: str, pos: Pos, out: Output) -> Tuple[Pos, Key]:
     pos += 1  # Skip "["
     pos = skip_chars(src, pos, TOML_WS)
     pos, key = parse_key(src, pos)
 
-    if state.flags.is_(key, Flags.EXPLICIT_NEST) or state.flags.is_(key, Flags.FROZEN):
+    if out.flags.is_(key, Flags.EXPLICIT_NEST) or out.flags.is_(key, Flags.FROZEN):
         raise suffixed_err(src, pos, f"Can not declare {key} twice")
-    state.flags.set(key, Flags.EXPLICIT_NEST, recursive=False)
+    out.flags.set(key, Flags.EXPLICIT_NEST, recursive=False)
     try:
-        state.out.get_or_create_nest(key)
+        out.data.get_or_create_nest(key)
     except KeyError:
         raise suffixed_err(src, pos, "Can not overwrite a value")
-    state.header_namespace = key
 
-    if src[pos : pos + 1] != "]":
+    if not src.startswith("]", pos):
         raise suffixed_err(src, pos, 'Expected "]" at the end of a table declaration')
-    return pos + 1
+    return pos + 1, key
 
 
-def create_list_rule(src: str, pos: Pos, state: State) -> Pos:
+def create_list_rule(src: str, pos: Pos, out: Output) -> Tuple[Pos, Key]:
     pos += 2  # Skip "[["
     pos = skip_chars(src, pos, TOML_WS)
     pos, key = parse_key(src, pos)
 
-    if state.flags.is_(key, Flags.FROZEN):
+    if out.flags.is_(key, Flags.FROZEN):
         raise suffixed_err(src, pos, f"Can not mutate immutable namespace {key}")
     # Free the namespace now that it points to another empty list item...
-    state.flags.unset_all(key)
+    out.flags.unset_all(key)
     # ...but this key precisely is still prohibited from table declaration
-    state.flags.set(key, Flags.EXPLICIT_NEST, recursive=False)
+    out.flags.set(key, Flags.EXPLICIT_NEST, recursive=False)
     try:
-        state.out.append_nest_to_list(key)
+        out.data.append_nest_to_list(key)
     except KeyError:
         raise suffixed_err(src, pos, "Can not overwrite a value")
-    state.header_namespace = key
 
-    end_marker = src[pos : pos + 2]
-    if end_marker != "]]":
-        raise suffixed_err(
-            src,
-            pos,
-            f'Found "{end_marker!r}" at the end of an array declaration.'
-            ' Expected "]]"',
-        )
-    return pos + 2
+    if not src.startswith("]]", pos):
+        raise suffixed_err(src, pos, 'Expected "]]" at the end of an array declaration')
+    return pos + 2, key
 
 
-def key_value_rule(src: str, pos: Pos, state: State, parse_float: ParseFloat) -> Pos:
+def key_value_rule(
+    src: str, pos: Pos, out: Output, header: Key, parse_float: ParseFloat
+) -> Pos:
     pos, key, value = parse_key_value_pair(src, pos, parse_float)
     key_parent, key_stem = key[:-1], key[-1]
-    abs_key_parent = state.header_namespace + key_parent
+    abs_key_parent = header + key_parent
 
-    if state.flags.is_(abs_key_parent, Flags.FROZEN):
+    if out.flags.is_(abs_key_parent, Flags.FROZEN):
         raise suffixed_err(
             src, pos, f"Can not mutate immutable namespace {abs_key_parent}"
         )
     # Containers in the relative path can't be opened with the table syntax after this
-    state.flags.set_for_relative_key(state.header_namespace, key, Flags.EXPLICIT_NEST)
+    out.flags.set_for_relative_key(header, key, Flags.EXPLICIT_NEST)
     try:
-        nest = state.out.get_or_create_nest(abs_key_parent)
+        nest = out.data.get_or_create_nest(abs_key_parent)
     except KeyError:
         raise suffixed_err(src, pos, "Can not overwrite a value")
     if key_stem in nest:
         raise suffixed_err(src, pos, "Can not overwrite a value")
     # Mark inline table and array namespaces recursively immutable
     if isinstance(value, (dict, list)):
-        abs_key = state.header_namespace + key
-        state.flags.set(abs_key, Flags.FROZEN, recursive=True)
+        out.flags.set(header + key, Flags.FROZEN, recursive=True)
     nest[key_stem] = value
     return pos
 
@@ -385,7 +376,7 @@ def parse_key_value_pair(
 
 def parse_key(src: str, pos: Pos) -> Tuple[Pos, Key]:
     pos, key_part = parse_key_part(src, pos)
-    key = [key_part]
+    key: Key = (key_part,)
     pos = skip_chars(src, pos, TOML_WS)
     while True:
         try:
@@ -393,11 +384,11 @@ def parse_key(src: str, pos: Pos) -> Tuple[Pos, Key]:
         except IndexError:
             char = None
         if char != ".":
-            return pos, tuple(key)
+            return pos, key
         pos += 1
         pos = skip_chars(src, pos, TOML_WS)
         pos, key_part = parse_key_part(src, pos)
-        key.append(key_part)
+        key += (key_part,)
         pos = skip_chars(src, pos, TOML_WS)
 
 
@@ -427,7 +418,7 @@ def parse_array(src: str, pos: Pos, parse_float: ParseFloat) -> Tuple[Pos, list]
     array: list = []
 
     pos = skip_comments_and_array_ws(src, pos)
-    if src[pos : pos + 1] == "]":
+    if src.startswith("]", pos):
         return pos + 1, array
     while True:
         pos, val = parse_value(src, pos, parse_float)
@@ -442,7 +433,7 @@ def parse_array(src: str, pos: Pos, parse_float: ParseFloat) -> Tuple[Pos, list]
         pos += 1
 
         pos = skip_comments_and_array_ws(src, pos)
-        if src[pos : pos + 1] == "]":
+        if src.startswith("]", pos):
             return pos + 1, array
 
 
@@ -452,7 +443,7 @@ def parse_inline_table(src: str, pos: Pos, parse_float: ParseFloat) -> Tuple[Pos
     flags = Flags()
 
     pos = skip_chars(src, pos, TOML_WS)
-    if src[pos : pos + 1] == "}":
+    if src.startswith("}", pos):
         return pos + 1, nested_dict.dict
     while True:
         pos, key, value = parse_key_value_pair(src, pos, parse_float)
@@ -478,7 +469,7 @@ def parse_inline_table(src: str, pos: Pos, parse_float: ParseFloat) -> Tuple[Pos
         pos = skip_chars(src, pos, TOML_WS)
 
 
-def parse_basic_str_escape(
+def parse_basic_str_escape(  # noqa: C901
     src: str, pos: Pos, *, multiline: bool = False
 ) -> Tuple[Pos, str]:
     escape_id = src[pos : pos + 2]
@@ -488,8 +479,9 @@ def parse_basic_str_escape(
         # the doc. Error if non-whitespace is found before newline.
         if escape_id != "\\\n":
             pos = skip_chars(src, pos, TOML_WS)
-            char = src[pos : pos + 1]
-            if not char:
+            try:
+                char = src[pos]
+            except IndexError:
                 return pos, ""
             if char != "\n":
                 raise suffixed_err(src, pos, 'Unescaped "\\" in a string')
@@ -514,7 +506,7 @@ def parse_basic_str_escape_multiline(src: str, pos: Pos) -> Tuple[Pos, str]:
 
 def parse_hex_char(src: str, pos: Pos, hex_len: int) -> Tuple[Pos, str]:
     hex_str = src[pos : pos + hex_len]
-    if len(hex_str) != hex_len or any(c not in string.hexdigits for c in hex_str):
+    if len(hex_str) != hex_len or not HEXDIGIT_CHARS.issuperset(hex_str):
         raise suffixed_err(src, pos, "Invalid hex value")
     pos += hex_len
     hex_int = int(hex_str, 16)
@@ -534,7 +526,7 @@ def parse_literal_str(src: str, pos: Pos) -> Tuple[Pos, str]:
 
 def parse_multiline_str(src: str, pos: Pos, *, literal: bool) -> Tuple[Pos, str]:
     pos += 3
-    if src[pos : pos + 1] == "\n":
+    if src.startswith("\n", pos):
         pos += 1
 
     if literal:
@@ -554,10 +546,10 @@ def parse_multiline_str(src: str, pos: Pos, *, literal: bool) -> Tuple[Pos, str]
 
     # Add at maximum two extra apostrophes/quotes if the end sequence
     # is 4 or 5 chars long instead of just 3.
-    if src[pos : pos + 1] != delim:
+    if not src.startswith(delim, pos):
         return pos, result
     pos += 1
-    if src[pos : pos + 1] != delim:
+    if not src.startswith(delim, pos):
         return pos, result + delim
     pos += 1
     return pos, result + (delim * 2)
@@ -580,7 +572,7 @@ def parse_basic_str(src: str, pos: Pos, *, multiline: bool) -> Tuple[Pos, str]:
         if char == '"':
             if not multiline:
                 return pos + 1, result + src[start_pos:pos]
-            if src[pos + 1 : pos + 3] == '""':
+            if src.startswith('"""', pos):
                 return pos + 3, result + src[start_pos:pos]
             pos += 1
             continue
@@ -595,13 +587,6 @@ def parse_basic_str(src: str, pos: Pos, *, multiline: bool) -> Tuple[Pos, str]:
         pos += 1
 
 
-def parse_regex(src: str, pos: Pos, regex: "Pattern") -> Tuple[Pos, str]:
-    match = regex.match(src, pos)
-    if not match:
-        raise suffixed_err(src, pos, "Unexpected sequence")
-    return match.end(), match.group()
-
-
 def parse_value(  # noqa: C901
     src: str, pos: Pos, parse_float: ParseFloat
 ) -> Tuple[Pos, Any]:
@@ -612,22 +597,22 @@ def parse_value(  # noqa: C901
 
     # Basic strings
     if char == '"':
-        if src[pos + 1 : pos + 3] == '""':
+        if src.startswith('"""', pos):
             return parse_multiline_str(src, pos, literal=False)
         return parse_one_line_basic_str(src, pos)
 
     # Literal strings
     if char == "'":
-        if src[pos + 1 : pos + 3] == "''":
+        if src.startswith("'''", pos):
             return parse_multiline_str(src, pos, literal=True)
         return parse_literal_str(src, pos)
 
     # Booleans
     if char == "t":
-        if src[pos + 1 : pos + 4] == "rue":
+        if src.startswith("true", pos):
             return pos + 4, True
     if char == "f":
-        if src[pos + 1 : pos + 5] == "alse":
+        if src.startswith("false", pos):
             return pos + 5, False
 
     # Dates and times
@@ -642,23 +627,9 @@ def parse_value(  # noqa: C901
     if localtime_match:
         return localtime_match.end(), match_to_localtime(localtime_match)
 
-    # Non-decimal integers
-    if char == "0":
-        second_char = src[pos + 1 : pos + 2]
-        if second_char == "x":
-            pos, hex_str = parse_regex(src, pos + 2, RE_HEX)
-            return pos, int(hex_str, 16)
-        if second_char == "o":
-            pos, oct_str = parse_regex(src, pos + 2, RE_OCT)
-            return pos, int(oct_str, 8)
-        if second_char == "b":
-            pos, bin_str = parse_regex(src, pos + 2, RE_BIN)
-            return pos, int(bin_str, 2)
-
-    # Decimal integers and "normal" floats.
+    # Integers and "normal" floats.
     # The regex will greedily match any type starting with a decimal
-    # char, so needs to be located after handling of non-decimal ints,
-    # and dates and times.
+    # char, so needs to be located after handling of dates and times.
     number_match = RE_NUMBER.match(src, pos)
     if number_match:
         return number_match.end(), match_to_number(number_match, parse_float)
